@@ -1,3 +1,4 @@
+/* vim: set tabstop=4: */
 /*
  * This file is part of TraceProto.
  * Copyright 2004 Eric Hope
@@ -35,12 +36,22 @@
 
 #include "traceproto.h"
 #include "config.h"
+#include "as.h"
+
+#ifdef HAVE_LIBDMALLOC
+#include <dmalloc.h>
+#endif /* HAVE_LIBDMALLOC */
 
 int main( int argc, char * argv[] )
 {
 
 	int cmd_arg, i;
 	char * opt_string;
+	char * env_string;
+
+#ifdef HAVE_LIBCAP
+	dropexcesscapabilities();
+#endif /* HAVE_LIBCAP */
 
 /*
  * regularize the program name
@@ -72,16 +83,22 @@ int main( int argc, char * argv[] )
 	behavior.do_audit = NO;
 	behavior.do_audit_exit = NO;
 	behavior.do_skip = NO;
+	behavior.as_discovery = NO;
 	behavior.output_style = TP_STD_OUTPUT;
 	behavior.payload_size = 12;
 	behavior.libnet_resolve_choice = LIBNET_RESOLVE;
 	behavior.hop_incr_unit = 1;
+	behavior.timestamp = NO;
+	behavior.timestamp_style = TP_TIMESTAMP_STD;
 	behavior.filter_text = "( host %s and dst port %d and src port %d ) or ( icmp[12:2] == %d )";
+	behavior.default_if = YES;
+	strncpy ( behavior.interface, "any", TP_IF_ARRAY - 1 ); 
 
 	packet.protocol_number = IPPROTO_TCP;
 	packet.payload = 0;
 	packet.frag_bit = 0;
 	packet.tcp_flags = TH_SYN;
+	packet.ip_packet_len = LIBNET_IPV4_H;
 
 	state.tcp_h  = LIBNET_PTAG_INITIALIZER;
 	state.udp_h  = LIBNET_PTAG_INITIALIZER;
@@ -97,7 +114,7 @@ int main( int argc, char * argv[] )
 	 * are a matter of different command line flags for the original
 	 * features.  We can more or less produce the same output behavior. 
 	 */
-	if ( strncmp ( state.prog, "traceroute", 10 ) == 0 )
+	if ( strncmp ( state.prog, "traceroute", 11 ) == 0 )
 	{
 		behavior.output_style = TP_CLASSIC_OUTPUT;
 		behavior.protocol = "udp";
@@ -107,11 +124,102 @@ int main( int argc, char * argv[] )
 		behavior.dst_port_incr = 1;	// dst port
 	}
 
+/*
+ * grab the relavent env variables
+ *
+ * NOTE: the values gathered from the env
+ * overwrite the defaults set above and are 
+ * overwritten in turn by any cmd line flags 
+ * that happen to be set below
+ *
+ * These should always have and else or a default so that the
+ * prog still does rational things if the user puts in a weird
+ * value.
+ */
+
+	env_string = NULL;
+	if ( ( env_string = getenv ( "TP_TIMESTAMP_STYLE" ) ) != NULL )
+	{
+		if ( strncmp ( env_string, "std", 3 ) == 0 )
+			behavior.timestamp_style = TP_TIMESTAMP_STD;
+		else if ( strncmp ( env_string, "us", 2 ) == 0 )
+			behavior.timestamp_style = TP_TIMESTAMP_US;
+		else if ( strncmp ( env_string, "desc", 4 ) == 0 )
+			behavior.timestamp_style = TP_TIMESTAMP_DESCEND;
+		else if ( strncmp ( env_string, "epoc", 4 ) == 0 )
+			behavior.timestamp_style = TP_TIMESTAMP_EPOC;
+		else
+			behavior.timestamp_style = TP_TIMESTAMP_STD;
+	}
+	
+	env_string = NULL;
+	if ( ( env_string = getenv ( "TP_DEFAULT_IF" ) ) != NULL )
+	{
+		strncpy ( behavior.interface, env_string, TP_IF_ARRAY - 1 );
+		behavior.default_if = NO;
+	}
+
+	env_string = NULL;
+	if ( ( env_string = getenv ( "TP_OUTPUT_STYLE" ) ) != NULL )
+	{
+		if ( 
+			strncmp ( env_string, "stand", 5 ) == 0 ||
+			strncmp ( env_string, "std",   3 ) == 0 ||
+			strncmp ( env_string, "s",     1 ) == 0    )
+		{
+				behavior.output_style = TP_STD_OUTPUT;
+				behavior.report = report_std;
+		} else if (	
+			strncmp ( env_string, "graph", 5 ) == 0 ||
+			strncmp ( env_string, "g",     1 ) == 0    )
+		{
+				behavior.output_style = TP_GRAPHIC_OUTPUT;
+				behavior.report = report_graphic;
+		} else if (
+			strncmp ( env_string, "class", 5 ) == 0 ||
+			strncmp ( env_string, "c",     1 ) == 0    )
+		{
+				behavior.output_style = TP_CLASSIC_OUTPUT;
+				behavior.report = report_classic;
+		} else if (
+			strncmp ( env_string, "no",    2 ) == 0 ||
+			strncmp ( env_string, "n",     1 ) == 0    )
+		{
+				behavior.output_style = TP_NO_OUTPUT;
+				behavior.report = report_none;
+		} else if (
+			strncmp ( env_string, "min",   3 ) == 0 ||
+			strncmp ( env_string, "m",     1 ) == 0    )
+		{
+				behavior.output_style = TP_MIN_OUTPUT;
+				behavior.report = report_minimum;
+		} else if ( 
+			strncmp ( env_string, "script", 6 ) == 0 ||
+			strncmp ( env_string, "p",      1 ) == 0    )
+		{
+				behavior.output_style = TP_SCRIPT_OUTPUT;
+				behavior.report = report_scriptable;
+		} else if ( 
+			strncmp ( env_string, "curs",  4 ) == 0 ||
+			strncmp ( env_string, "ncurs", 5 ) == 0 ||
+			strncmp ( env_string, "C",     1 ) == 0    )
+		{
+				behavior.output_style = TP_CURSES_OUTPUT;
+				behavior.report = report_curses;
+				tixe_cleanup.curses_end = YES;
+		} else {
+				behavior.output_style = TP_STD_OUTPUT;
+				behavior.report = report_std;
+		}
+	} // end if getenv ( "TP_OUTPUT_STYLE" )
+	
+
+
 /*
  * get command line args
  */
 	while ( ( cmd_arg = getopt ( argc, argv,
-			"cCvnhRfp:i:I:D:r:t:k:o:a:s:S:H:d:M:m:w:W:p:P:"
+			"cCvnhRfATp:i:I:D:r:t:k:o:a:s:S:H:d:M:m:w:W:F:p:P:z:"
 		) ) != EOF )
 	{
 		switch ( cmd_arg )
@@ -233,6 +341,10 @@ int main( int argc, char * argv[] )
 				behavior.output_style = TP_CLASSIC_OUTPUT;
 				behavior.report = report_classic;
 				break;
+			} else if ( opt_string [ 0 ] == 'C' ) {
+				behavior.output_style = TP_CURSES_OUTPUT;
+				behavior.report = report_curses;
+				break;
 			} else if ( opt_string [ 0 ] == 'n' ) {
 				behavior.output_style = TP_NO_OUTPUT;
 				behavior.report = report_none;
@@ -262,11 +374,26 @@ int main( int argc, char * argv[] )
 		case 'f':
 			packet.frag_bit = TP_DONT_FRAG;
 			break;
+		case 'F':
+			strncpy ( behavior.interface, optarg, TP_IF_ARRAY - 1 );
+			behavior.default_if = NO;
+			break;
+		case 'A':
+			behavior.as_discovery = YES;
+			break;
+		case 'T':
+			behavior.timestamp = YES;
+			make_timestamp ( behavior.timestamp_str );
+			break;
 		case 'R':
 			/* start at the distant end and trace consecutively
 			 * closer until we get here */
 			behavior.hop_incr_unit = -1;
 			break;
+		case 'z':
+			/* undocumented dubug flag */
+			parse_debug ( optarg );
+			break;
 		default:
 			usage( state.prog );
 			tixe ( tixe_cleanup, 1 );
@@ -351,16 +478,20 @@ int main( int argc, char * argv[] )
 	{
 		behavior.protocol = "udp";
 		packet.protocol_number = IPPROTO_UDP;
+		packet.ip_packet_len += LIBNET_UDP_H;
 	}
 	else if ( strncmp( behavior.protocol, "tcp", 3 ) == 0 )
 	{
 		behavior.protocol = "tcp";
 		packet.protocol_number = IPPROTO_TCP;
+		packet.ip_packet_len += LIBNET_TCP_H;
 	}
 	else if ( strncmp( behavior.protocol, "icm", 3 ) == 0 )
 	{
 		behavior.protocol = "icmp";
 		packet.protocol_number = IPPROTO_ICMP;
+		behavior.filter_text = "( icmp[12:2] == %d ) or ( host %s and icmp[0] == 0 )";
+		packet.ip_packet_len += LIBNET_ICMPV4_ECHO_H;
 	}
 	else
 	{
@@ -453,6 +584,13 @@ int main( int argc, char * argv[] )
 		tixe ( tixe_cleanup, 1 );
 	}
 
+	/*
+	 * Explicitly turn off continuous_accounting with the curses
+	 * It won't work well and is redundent.
+	 */
+	if ( behavior.output_style == TP_CURSES_OUTPUT )
+		behavior.continuous_accounting = NO;
+
 	/*
 	 * note that the default payload is 12 bytes due to a udp
 	 * peculiarity.  less than 12 bytes and no udp response
@@ -475,8 +613,22 @@ int main( int argc, char * argv[] )
 		}
 		tixe_cleanup.payload_free = YES;
 		memset ( packet.payload, '\0', behavior.payload_size );
+
+		packet.ip_packet_len += behavior.payload_size;
 	}
 
+/*
+ * Setup the free list for the packet alignment func
+ */
+	tp_align_freelist.next = ( struct tp_align_ref * ) malloc ( sizeof ( struct tp_align_ref ) );
+	if ( tp_align_freelist.next == NULL )
+	{
+		printf ( "error allocating memory for the free list\n" );
+		tixe (  tixe_cleanup, 1 );
+	}
+	tixe_cleanup.free_list_free = YES;
+
+
 /*
  * DNS stuff
  */
@@ -515,11 +667,23 @@ int main( int argc, char * argv[] )
 	tixe_cleanup.hop_record_free = YES;
 
 /*
- * start packet
+ * find interface if needed and start packet
  */
+	//find_interface ( );
+
+	if ( debug.interface == YES )
+	{
+		printf ( "debug: using default interface = %s\n",
+			behavior.default_if ? "YES"  : "NO" );
+		printf ( "debug: interface flag is %s\n",
+			behavior.default_if == NO ? behavior.interface : "NULL" );
+		
+	}
+
 	state.packet = libnet_init (
 		LIBNET_RAW4,	    /* inject type */
-		NULL,		   /* device */
+	//	NULL,		   /* device */
+		behavior.default_if == NO ? behavior.interface : NULL,
 		(u_char *) state.error_buff );
 	if (!state.packet)
 	{
@@ -551,7 +715,10 @@ int main( int argc, char * argv[] )
 /*
  * do the pcap stuff and the initial filter
  */
-	state.psocket = pcap_open_live( "any",
+	//state.psocket = pcap_open_live( NULL,
+	//state.psocket = pcap_open_live( behavior.interface,
+	state.psocket = pcap_open_live (
+				behavior.default_if == NO ? behavior.interface : NULL,
 				SNAPLEN,
 				NO_PROMISC,
 				( behavior.wait_timeout * 1000 ),
@@ -613,6 +780,15 @@ int main( int argc, char * argv[] )
 		tixe ( tixe_cleanup, 1 );
 	}
 
+/*
+ * if we're doing as_discovery, set it up here
+ */
+	if ( behavior.as_discovery == YES && setup_as ( ) != 0 )
+	{
+		printf ( "error doing AS discovery: %s\n", as_string );
+		behavior.as_discovery = NO;
+	}
+
 /*
  * set up the signals
  */
@@ -628,7 +804,7 @@ int main( int argc, char * argv[] )
 		state.current_hop = behavior.max_ttl;
 /*
  * MAIN LOOP STARTS HERE
-
+ *
  * loop through, building the ip
  * and sending it
  */
@@ -680,11 +856,11 @@ int main( int argc, char * argv[] )
 
 /*
  * build the ip
- * (this should probably be rolled in build_packet() )
+ * ( this should probably be rolled in build_packet() )
  */
 		packet.ip_id = libnet_get_prand(LIBNET_PRu16);
 		state.ip_h = libnet_build_ipv4(
-			LIBNET_IPV4_H,			/* IP hdr len */
+			packet.ip_packet_len,	/* IP packet len */
 			0,				/* tos */
 			packet.ip_id,			/* frag id */
 			packet.frag_bit,		/* frag offset */
@@ -697,6 +873,20 @@ int main( int argc, char * argv[] )
 			0,				/* payload size */
 			state.packet,
 			state.ip_h );
+
+		if ( debug.packet_length == YES )
+			printf ( "\nHEADER_LENGTH IP:%d payload:%d, total:%d\n",
+				LIBNET_IPV4_H, behavior.payload_size, packet.ip_packet_len );
+
+		if ( debug.send_buf == YES )
+		{
+			printf ( "debug: send buffer:\n" );
+			debug_packet ( libnet_getpbuf ( state.packet, state.ip_h ),
+				  libnet_getpbuf_size ( state.packet, state.ip_h ) );
+			debug_packet ( libnet_getpbuf ( state.packet, state.tcp_h ),
+				  libnet_getpbuf_size ( state.packet, state.tcp_h ) );
+			debug_packet ( packet.payload, behavior.payload_size );
+		}
 /*
  * and change the filter
  */
@@ -714,6 +904,8 @@ int main( int argc, char * argv[] )
 
 		gettimeofday ( & state.start_time, NULL );
 
+		if ( debug.loop == YES ) { printf ( "debug: time checked, writing packet\n" ); }
+
 		if ( libnet_write( state.packet ) == -1 )
 		{
 			printf("error at send: %s\n",
@@ -721,6 +913,8 @@ int main( int argc, char * argv[] )
 			tixe ( tixe_cleanup, 1 );
 		}
 
+		if ( debug.loop == YES ) { printf ( "debug: packet written, waiting\n" ); }
+
 /*
  * wait for a packet to get
  * through the filter, timeout
@@ -733,6 +927,8 @@ int main( int argc, char * argv[] )
 			FD_ZERO( & ( state.wheel ) );
 			FD_SET( state.fake_psocket->fd, & ( state.wheel ) );
 
+			if ( debug.loop == YES ) { printf ( "debug: select\n" ); }
+
 			if ( select( ( state.fake_psocket->fd ) + 1,
 					& ( state.wheel ),
 					NULL,
@@ -745,16 +941,27 @@ int main( int argc, char * argv[] )
 				 */
 				gettimeofday ( & ( state.end_time ), NULL );
 
+				if ( debug.loop == YES )
+					printf ( "debug: select returned, capturing\n" );
+
 				state.capture_buf = pcap_next(
 						state.psocket,
 						& ( state.psock_hdr ) );
 
+				if ( debug.loop == YES )
+					printf ( "debug: captured, parsing\n" );
+
 				state.trip_time = diff_time (
 					& ( state.start_time ),
 					& ( state.end_time ) );
 
 				if ( parse_packet( state.capture_buf ) > 0 )
-					account_packet ( state.trip_time );
+				{
+					freelist_cleaner ( );
+				}
+
+				if ( debug.loop == YES )
+					printf ( "debug: parsed, resets\n" );
 
 				if ( packet.protocol_number == IPPROTO_TCP
 					&& behavior.tcp_resets == YES
@@ -772,9 +979,12 @@ int main( int argc, char * argv[] )
 				/*
 				 * if we get here,  no one wanted to talk to us for this probe
 				 */
+				account_packet ( 0 );
 				behavior.report( TP_OUT_HOP_INFO, ( struct in_addr * ) NULL, TP_TYPE_NR );
 				state.packet_match = TP_PACKET_TIMEOUT;
-				account_packet ( 0 );
+
+				if ( debug.loop == YES )
+					printf ( "debug: no answer (%d)\n", FD_ISSET ( state.fake_psocket->fd, & ( state.wheel ) ) );
 			}
 
 			/*
@@ -782,7 +992,10 @@ int main( int argc, char * argv[] )
 			 */
 			if ( behavior.do_audit == YES || behavior.do_audit_exit == YES )
 			{
-				hop_audit();
+				if ( behavior.do_audit_exit == YES )
+					behavior.report( TP_OUT_FOOTER, NULL, 0 );
+				else
+					hop_audit();
 			}
 		} // end of while ( packet_match == TP_PACKET_NO ) loop
 
@@ -869,6 +1082,8 @@ int main( int argc, char * argv[] )
 		else if ( packet.dst_port <= behavior.min_dst_port && behavior.dst_port_incr == -1)
 			packet.dst_port = behavior.max_dst_port + 1;
 
+		if ( debug.loop == YES ) { printf ( "debug, looping back to the beginning\n" ); }
+
 	} // end of the main send/recv loop
 
 /*
@@ -877,7 +1092,8 @@ int main( int argc, char * argv[] )
  * spit out the stats, sweep up after ourselves, and close up shop
  * last packet out please shut off the lights
  */
-	hop_audit();
+	behavior.report ( TP_OUT_FOOTER, NULL, 0 );
+	//hop_audit();
 
 	tixe ( tixe_cleanup, 0 );	// cleanup and exit, status 0
 
